/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#ifdef CLIENT
#include "client.h"
#include <string.h>

#ifdef USE_T1LIB
# include <t1lib.h>
#endif

/* they are automatically sorted by time */
struct message {
	unsigned char	*data;
	unsigned long	timestamp;
	struct message	*next;
};
typedef struct message MESSAGE;

static MESSAGE 		*firstmessage;
static int		nmessages;

void setmessagetime ( unsigned long time ) {
	prefs->messagetime = time;
}

/* from==-1 is itself, 0 is server, else number of client */
void addmessage ( unsigned char *m, unsigned short from ) {
	MESSAGE		*tmp, *new;
	char		who[10];

	if ( !m ) {
		fprintf( stderr, "addmessage(): null message\n" );
		return;
	}

	if ( !(new = (MESSAGE *)malloc(sizeof(MESSAGE))) ) {
		fprintf( stderr, "Couldn't allocate memory in addmessage()\n" );
		return;
	}
	
	if ( from == 0 )
		sprintf( who, "S: " );
	else if ( from > 0 )
		sprintf( who, "%d: ", from );

	new->data = (unsigned char *)malloc(strlen(who)+strlen(m)+1);
	strcat( new->data, who );
	strcat( new->data, m );
	new->timestamp = timestamp( NULL );
	
	if ( firstmessage == NULL )
		firstmessage = new;
	else {
		for ( tmp = firstmessage; tmp->next != NULL; tmp = tmp->next ) ;
		tmp->next = new;
	}
	new->next = NULL;
	nmessages++;
}

void freemessage ( MESSAGE *m ) {
	MESSAGE 	*tmp;

	if ( firstmessage == NULL )
		return;
	else if ( firstmessage == m )
		firstmessage = m->next;
	else {
		for ( tmp = firstmessage; tmp->next != m; tmp = tmp->next ) ;
		tmp->next = m->next;
	}
	free( m->data );
	free( m );
	nmessages--;
}

void cleanmessages ( void ) {
	unsigned long 	timenow = timestamp( NULL );
	int 		i;
	
	if ( !firstmessage )
		return;

	/* clean by number */
	if ( nmessages > prefs->maxmessages )
		for ( i = nmessages - prefs->maxmessages; i > 0; i-- )
			free( firstmessage );

	/* clean by timestamp */
	while ( timenow - firstmessage->timestamp > prefs->messagetime )
			freemessage( firstmessage );
}

#ifdef USE_T1LIB

static int	fontid, pad;

void initFont ( void ) {
	nmessages = 0;
	fontid = 0;
	pad = 8;
	T1_SetBitmapPad( pad );
	T1_InitLib( NO_LOGFILE );
}

void closeFont ( void ) {
	T1_CloseLib();
}

void showmessages ( void ) {
	GLYPH 		*glyph;
	MESSAGE		*tmp;
	int		i, j, h, w, p;
	
	cleanmessages();
	if ( !firsmessage )
		return;
	
	if ( engprefs->aa == True )
		for ( tmp = firstmessage; tmp->next != NULL; tmp = tmp->next ) {
			glyph = T1_AASetString( fontid, tmp->data, 0, 0,
				 T1_KERNING, fontsize, NULL );
			/* now draw */
		}
	else 
		for ( tmp = firstmessage; tmp->next != NULL; tmp = tmp->next ) {
			glyph = T1_SetString( fontid, tmp->data, 0, 0,
				 T1_KERNING, fontsize, NULL );
			/* now draw. The lines below are adapted from T1 source */
			h = glyph->metrics.ascent-glyph->metrics.descent;
			w = glyph->metrics.rightSideBearing-glyph->metrics.leftSideBearing;
			p = PAD( w, pad )/pad;
			for ( i = 0; i < h*p; i += p )
				for ( j = 0; j < p; j++ )
					for ( k = 0; k < pad; k++ )
/* we are not handling wrapping! */
//						putPixel( (glyph->bits[i+j]>>k)&0x01 );
/* note: as the original code shows:
      bin_dump_c(glyph->bits[i*paddedW/T1_pad+j], 1);
      bin_dump_s(((unsigned short *)glyph->bits)[i*paddedW/T1_pad+j], 1);
      bin_dump_l(((unsigned long *)glyph->bits)[i*paddedW/T1_pad+j], 1);
we need to add the type modifiers, or at least a warning will be returned */
		}
}

#else /* not USE_T1LIB, use plain GGI functions */

static ggi_coord font;

void initFont ( void ) {
	nmessages = 0;
	ggiGetCharSize( vis, (int *)&font.x, (int *)&font.y );
}

void closeFont ( void ) {
}

/* this is pretty bad */
void showmessages ( void ) {
	MESSAGE 	*tmp;
	int 		i;

	for ( i = 0, tmp = firstmessage; tmp != NULL; tmp = tmp->next, i++ )
		ggiPuts( vis, 5, font.y*i, tmp->data );
}

#endif /* USE_T1LIB */
#endif /* CLIENT */
