/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "unlimited.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#define EARTH_GRAVITY		9.8		/* m/s^2 */
#define NORMAL_TEMPERATURE	300		/* K */

#define GENERALPREF		"general.pref"

GENERAL_UL general;

#define YES( c )	( c == 'y' || c == 'Y' )

/* for fields regarding other applications, like sound or video output, 0
if a numeric filed or '\0' if a string will set defaults */
/*** should detect errors with scanfs and etc */
int initGeneral ( void ) {
	FILE *input;
	char temp[512], *value;
	int i = 0;

	general.gravity = EARTH_GRAVITY;
	general.temperature = NORMAL_TEMPERATURE;
	general.useobjso = False;
	general.createobject = False;
	general.destroyobject = False;
	general.modifyobject = True;
	general.rigidobjects = False;
	
	sprintf( temp, "%s%s", ULDIR, GENERALPREF );
	if ( (input = fopen( temp, "r" )) ) {
		while ( fgets( temp, sizeof( temp ), input )) {
			i++;
			
			if ( (value = strchr( temp, '#' )) != NULL )
				*value = '\0';
			
			value = strchr( temp, '=' );
			if ( !value )
				continue;
			*value++ = '\0';

			while( isspace( *value ) )
				value++;

			if ( strstr( temp, "gravity" ) )
				general.gravity = atof( value );
			else if ( strstr( temp, "temperature" ) )
				general.temperature = atof( value );
			else if ( strstr( temp, "useobjso" ) )
				general.useobjso = ( YES(*value) ? True : False );
			else if ( strstr( temp, "createobject" ) )
				general.createobject = ( YES(*value) ? True : False );
			else if ( strstr( temp, "destroyobject" ) )
				general.destroyobject = ( YES(*value) ? True : False );
			else if ( strstr( temp, "modifyobject" ) )
				general.modifyobject = ( YES(*value) ? True : False );
			else if ( strstr( temp, "rigidobjects" ) )
				general.rigidobjects = ( YES(*value) ? True : False );
			else
				fprintf( stderr, "Syntax error in %s, line %d; ignoring\n",
					 GENERALPREF, i );
		}
		fclose( input );
		return OK;
	}
	else {
		fprintf( stderr, 
			"Could not open preferences file: %s. Trying to create a new one\n",
			GENERALPREF );
		saveGeneral();
		return OK;
	}
		
}

#define YESNO( d )		( d == True ? "yes" : "no" )

void saveGeneral ( void ) {
	unsigned char tmp[512];
	FILE *out;	

	sprintf( tmp, "%s%s", ULDIR, GENERALPREF );
	if ( (out = fopen( tmp, "w" )) ) {
		fprintf( out, "# Unlimited Simulator server preferences file\n" );
		fprintf( out, "gravity = %f\n", general.gravity );
		fprintf( out, "temperature = %f\n", general.temperature );
		fprintf( out, "useobjso = %s\n", YESNO( general.useobjso ) );
		fprintf( out, "createobject = %s\n", YESNO( general.createobject ) );
		fprintf( out, "destroyobject = %s\n", YESNO(general.destroyobject ) );
		fprintf( out, "modifyobject = %s\n", YESNO( general.modifyobject ) );
		fprintf( out, "rigidobjects = %s\n", YESNO( general.rigidobjects ) );
		fprintf( out, "# End of configuration file" );
		fclose( out );
	}
	else
		fprintf( stderr, "Could not open preferences file to save\n" );
}
