/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "unlimited.h"
#include "mathlib.h"
#include <dlfcn.h>

#define G		9.8	/* gravity */
#define AIRDENSITY	1.2	/* kg/m^3 */

/* ps() (physical simulator) calculates all forces and torques applied on the
   car; then it does all 3D mechanical calculations and the results are 
   returned in the car structure. Coordinates are in local (car) space, unless
   noted. Z is up, X to right, Y to front */
void ps ( CAR_UL *car, USERINPUT_UL *ui, double frametime ) {
	int i;
	VECTOR forces, torques, tempf, tempt;

	zeroVector( &result );
	if ( prefs.controller != COMPUTER)
		handleUI( ui );
	else
		car->ai( ui );
	
	/* calculate gear, rpm, engine thrust here, fuel burned,
	  take from mass */
	car->weight = car->mass*G;
	
	/* project weight to local system and add to result */
	/* calculate engine force */
	/* correct steer */
	/* calculate friction and other forces in each wheel; acceleration
	  here is dependent of engine! */
	/* pass them to suspension */

	car->airfriction( car, tempf, tempt );
	addV( forces, tempf, forces );
	addV( torques, tempt, torques );
	
	
}

static void handleUI ( USERINPUT_UL *ui ) {
	double value[3];
	/* by now, both equal. in future, joystick's proportional to the
	variation of the angle */
	/* missing: the joystick check for position, read value[] fro the 
	car specs: probably add a function to it */

	if ( prefs.controller == KEYBOARD )	value = { 0.05, 4.0, 10.0 }; 
	else if ( prefs.controller == JOYSTICK ) value = { 0.05, 4.0, 10.0 };
	
	if ( ui.steer.touched == True ) {
		if ( ui.steer.extra == 1 )
			if ( ui.steer.value + value[0] <= car->maxsteer ) {
				ui.steer.value += value[0];
				if ( ui.steer.value > car->maxsteer )
					ui.steer.value = car->maxsteer;
			}
		else
			if ( ui.steer.value - value[0] >= -car->maxsteer ) {
				ui.steer.value -= value[0];
				if ( ui.steer.value < -car->maxsteer )
					ui.steer.value = -car->maxsteer;
			}
	}
	else {
		if ( ui.steer.value > 0 )
			ui.steer.value -= value[0];
			if ( ui.steer.value < 0 )
				ui.steer.value = 0;
		else ( ui.steer.value < 0 ) 
			ui.steer.value += value[0];
			if ( ui.steer.value > 0 )
				ui.steer.value = 0;
	}
	
	if ( ui.accel.touched == True )
		if ( ui.thrust.value < 100.0 ) {
			ui.thrust.value += 4.0;
			if ( ui.thrust.value > 100.0 )
				ui.thrust.value = 100.0;
		}
		  	/* what about nonlinear? */
	else
		if ( ui.thrust.value > 0.0 ) {
			ui.thrust.value -= 4.0;
			if ( ui.thrust.value < 0.0 )
				ui.thrust.value = 0.0;
		}

	if ( ui.brake.touched == True )
		if ( ui.braking.value < 100.0 ) {
			ui.braking.value += 4.0;
			if ( ui.braking.value > 100.0 )
				ui.braking.value = 100.0;
		}
	else
		if ( ui.braking.value > 0.0 ) {
			ui.braking.value -= 4.0;
			if ( ui.braking.value < 0.0 )
				ui.braking.value = 0.0;
		}
		
	/* gears are handled on key reading */

	if ( ui.clutch.touched == True )
		if ( ui.clutch.value <= 100.0 ) {
			ui.clutch.value += 10.0;
			if ( ui.clutch.value > 100.0 )
				ui.clutch.value = 100.0;
		}
	else
		if ( ui.clutch.value > 0.0 ) {
			ui.clutch.value -= 10.0;
			if ( ui.clutch.value < 0.0 )
				ui.clutch.value = 0.0;
		}
}

void defAirFriction ( CAR_UL *car, VECTOR_UL *force, VECTOR_UL *torque ) {
	/* F = 1/2 * A * C * d * v^2 */
	
	if ( car->velocity.y )
		force->y = car->area * AIRDENSITY * car->drag *
			car->velocity.y * car->velocity.y / 2 );
	else ( force->y = 0 );
	
	force->x = force->z = 0;	
	
	zeroVector( torque );
}

void defWheelFriction ( WHEEL_UL *wheel, VECTOR_UL *force, VECTOR_UL *torque ) {
	double d;
	zeroVector( force );
	zeroVector( torque );
	d = wheel->normal->y * wheel->terrain->friction[weather];
	switch ( wheel->state ) {
		case OUTOFGROUND:
			break;
		case NORMAL:
			wheel->wearing -= 10 * frametime *
					wheel->terrain->friction[weather];
			force->y = 0.01 * COS(wheel->angle) * d;
			force->x = 0.01 * SIN(wheel->angle) * d;
			torque->z = force->x * wheel->dist.x;
			torque->x = force->y * wheel->dist.y;
			break;
		case SPINNING:
			wheel->wearing -= 1000 * frametime *
					wheel->terrain->friction[weather];
			break;
		case BURNING:
			wheel->wearing -= 1000 * frametime *
					wheel->terrain->friction[weather];
			break;
		case BLOWN:
			force->y = COS(wheel->angle) * d;
			force->x = SIN(wheel->angle) * d;
			break;
		case PUNCTURED:
			force->y = COS(wheel->angle) * d;
			force->x = SIN(wheel->angle) * d;
			break;

	}
	
}

/* engine's torque to the car is negligible... */
void defEngine ( CAR_UL *car, USERINPUT_UL *ui, VECTOR_UL *force ) {
	if ( car->gertype == MECHANICAL && ui->clutch > 0.1 ) {
		zeroVector( force );
		/* correct rpm */
		return;
	}
}

void defSuspension ( void ) {
}

enum {
	NOCHANGE, PARK, GEARUP, GEARDOWN, REVERSE
} gearps_ul;
typedef enum gearps_ul GEARPS_UL;

void defGear ( CAR_UL *car, USERINPUT_UL *ui ) {
	GEARPS_UL gear = NOCHANGE;
	if ( ui->clutch )
		gear = PARK;
	switch ( car->geartype ) {
		case AUTOMATIC:
			if ( car->rotation > 0.8*car->maxrotation && 
			     car->gear > 0 ) {
				car->gear--;
				gear = (car->gear == 0 ? PARK : GEARDOWN );
			}
			else if ( car->rotation < 0.2*car->maxrotation && 
			     car->gear < car->maxgear && car->gear > 0 ) {
				car->gear++;
				gear = GEARUP;
			}
			else if ( ui->gear == -1 && 
				  ( car->gear == 1 || car->gear == 0 ) ) {				
				car->gear--;
				gear = ( car->gear == 0 ? PARK : REVERSE );
			}
			else if ( ui->gear == 1 && 
				  ( car->gear == -1 || car->gear == 0 ) ) {
				car->gear++;
				gear = ( car->gear == 0 ? PARK : GEARUP );
			}
			break;
/*** done til here */
		case NOCLUTCH:
			if ( ui->gear == -1 && gear > -1 ) {
				car->gear--;
				if ( car->gear == 0 )
					gear = PARK;
				if ( car->gear == -1 )
					gear = REVERSE;
				else
					gear = GEARDOWN;
			}
			else if ( car->rotation < 0.2*car->maxrotation && 
			     car->gear < car->maxgear && car->gear > 0 ) {
				car->gear++;
				gear = GEARUP;
			}
			else if ( ui->gear == -1 && 
				  ( car->gear == 1 || car->gear == 0 ) ) {				
				car->gear--;
				gear = ( car->gear == 0 ? PARK : REVERSE );
			}
			else if ( ui->gear == 1 && 
				  ( car->gear == -1 || car->gear == 0 ) ) {
				car->gear++;
				gear = ( car->gear == 0 ? PARK : GEARUP );
			}
			break;
		case MECHANICAL:
			if ( ui->gear != 0 ) {
				/* deal with clutch */
				car->gear += ui->gear;
				if ( car->gear > car->maxgear )
					car->gear = car->maxgear;
				else if ( car->gear < -1 )
					car-> = -1;
				
				/* deal with RPM, etc */
			}
			break;
	}
	switch ( gear ) {
		case NOCHANGE:
			break;
		case REVERSE:
			
			break;
		case PARK:
			
			break;
		case GEARUP:
			
			break;
		case GEARDOWN:
			
			break;
	}
}
