/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#ifdef CLIENT

#include "ulengine.h"
#include <malloc.h>

#define MIN( a, b )		( a < b ? a : b )
#define MAX( a, b )		( a > b ? a : b )
#define MaxPoint( a,b ) 	( a->y > b->y ? a : b)
#define MinPoint( a,b )		( b->y > a->y ? a : b)
#define MaxPoint3(a,b,c)	MaxPoint( MaxPoint( a, b ), MaxPoint( b, c ) )
#define MinPoint3(a,b,c)	MinPoint( MinPoint( a, b ), MinPoint( b, c ) )

TRI_UL		ultri;

sInt initGraphic ( void ) {
	sInt i;
	ultri.npoints = 3;
	if( !(ultri.points = (POINT_UL *)malloc(3*sizeof(POINT_UL))) ) {
		fprintf( stderr, "initGraphic, 1\n" );
		return FATALERROR;
	}
	for ( i = 0; i < ultri.npoints; i++ ) {
		if ( !(ultri.points[i].vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL)))) {
			fprintf( stderr, "initGraphic, 2\n" );
			return FATALERROR;
		}
		if ( !(ultri.points[i].norm = (VECTOR_UL *)malloc(sizeof(VECTOR_UL)))) {
			fprintf( stderr, "initGraphic, 3\n" );
			return FATALERROR;
		}
	}
	if ( !(ultri.color = (ggi_color *)malloc( ultri.npoints
			* sizeof(ggi_color))) ) {
		fprintf( stderr, "initGraphic, 4\n" );
		return FATALERROR;
	}
	if ( !(ultri.flatcolor = (ggi_color *)malloc(sizeof(ggi_color))) ) {
		fprintf( stderr, "initGraphic, 5\n" );
		return FATALERROR;
	}
	ultri.visible = True;
	return OK;
}

void closeGraphic ( void ) {
	sInt 		i;

	free( ultri.color );
	free( ultri.flatcolor );
	for ( i = 0; i < ultri.npoints; i++ ) {
		free( ultri.points[i].vect );
		free( ultri.points[i].norm );
	}
	free( ultri.points );
}

void copyColor( ggi_color *target, ggi_color *src ) {
	memcpy( target, src, sizeof(ggi_color) );
}

ggi_color RGBAtoColor ( uShort r, uShort g, uShort b, uShort a ) {
	ggi_color c;
	c.r = r;	 c.g = g; 	c.b = b;	 c.a = a;
	return c;
}

/* this funciton is supposed to deal with buffers, no-buffers, etc. */
void putPixel ( sInt x, sInt y, ggi_color *color ) {
	uLong	 	l;
	
	if ( db == False ) {
		ggiPutPixel( vis, x, y, ggiMapColor( vis, color ));
	}
	else {
		l = (color->r<<16) | (color->g<<8) | (color->b); 
		/* this line works here */
	/* this one should work:
		memcpy( im[framenum].fbptr + y*im[framenum].stride+x*(im[framenum].pix->size>>3), l, 
			(im[framenum].pix->size>>3) ); */
	}
}

void getPixel ( sInt x, sInt y, ggi_color *color ) {
	ggi_pixel	pix;

	if ( db == False ) {
		ggiGetPixel( vis, x, y, &pix );
		ggiUnmapPixel( vis, pix, color );
	}
	else { /* direct buffers */
		
	}
}

void drawPixelD ( dFloat x, dFloat y, dFloat z, ggi_color *color ) {
	sInt	ix, iy;

	ix = round(x);
	iy = round(y);

	if ( !engpref.aa ) {
		if ( z < frame.zbuf[iy][ix] || frame.zbuf[iy][ix] == 0 ) { 
			copyColor( &frame.color[iy][ix], color);
			frame.zbuf[iy][ix] = (fFloat)z;
		}
	}
	else { /* antialiased */
		
	}
}

/* revised, working 100% but slow due to individual calls to drawPixel! 
Missing anti aliasing support.  */
void drawHLine ( dFloat x1, dFloat x2, dFloat y, dFloat z1, dFloat z2, 
		ggi_color *c1, ggi_color *c2 ) {
	fFloat		nx1, nx2, dx, rx1, rx2;
	fFloat		zm, rm = 0, gm = 0, bm = 0, am = 0, nz;
	ggi_color	c;

	rx1 = round(x1);
	rx2 = round(x2);
	if ( rx1 == rx2 ) {
		c = RGBAtoColor((c1->r+c2->r)>>1, (c1->g+c2->g)>>1,
				(c1->b+c2->b)>>1, (c1->a+c2->a)>>1 );
		drawPixelD( x1, y, (z1+z2)/2, &c );
		return;
	}
	else if ( rx1 < rx2 ) {
		nx1 = rx1;		nx2 = rx2;
		nz = z1;		dx = 1/(nx2-nx1+1);
		zm = (z2-z1)*dx;
		copyColor( &c, c1 );
		if ( c1 != c2 ) {
			rm = ((fFloat)(c2->r - c.r))*dx;
			gm = ((fFloat)(c2->g - c.g))*dx;
			bm = ((fFloat)(c2->b - c.b))*dx;
			am = ((fFloat)(c2->a - c.a))*dx;
		}
	}
	else {
		nx1 = rx2;		nx2 = rx1;
		nz = z2;		dx = 1/(nx2-nx1+1);
		zm = (z1-z2)*dx;
		copyColor( &c, c2 );
		if ( c1 != c2 ) {
			rm = ((fFloat)(c1->r - c.r))*dx;
			gm = ((fFloat)(c1->g - c.g))*dx;
			bm = ((fFloat)(c1->b - c.b))*dx;
			am = ((fFloat)(c1->a - c.a))*dx;
		}
	}
	
	for ( ; nx1 <= nx2; nx1++, nz += zm, c.r += rm, c.g += gm, 
			c.b += bm, c.a += am ) {
		drawPixelD( nx1, y, nz, &c );
	}
}

/* draw a horizontal line interpoling from a texture. */
void drawHLineTex ( dFloat x1, dFloat x2, dFloat y, dFloat z1, dFloat z2, 
		IMAGE_UL *image, fFloat u1, fFloat u2, fFloat v1, fFloat v2 ) {
	fFloat		nx1, nx2, dx, rx1, rx2;
	fFloat		zm, um, vm, nz, nu, nv;
	ggi_color	c;

	rx1 = round(x1);
	rx2 = round(x2);
	if ( rx1 == rx2 ) {
		c = pixelImage( image, u1, v1 );
		drawPixelD( x1, y, (z1+z2)/2, &c );
		return;
	}
	else if ( rx1 < rx2 ) {
		nx1 = rx1;		nx2 = rx2;
		nz = z1;		dx = 1/(nx2-nx1+1);
		nu = u1;		nv = v1;
		zm = (z2-z1)*dx;
		um = (u2-u1)*dx;
		vm = (v2-v1)*dx;
		c = pixelImage( image, u1, v1 );
	}
	else {
		nx1 = rx2;		nx2 = rx1;
		nz = z2;		dx = 1/(nx2-nx1+1);
		nu = u2;		nv = v2;
		zm = (z1-z2)*dx;
		um = (u1-u2)*dx;
		vm = (v1-v2)*dx;
		c = pixelImage( image, u2, v2 );
	}
	
	for ( ; nx1 <= nx2; nx1++, nz += zm, nu += um, nv += vm,
			c = pixelImage( image, nu, nv ) ) {
		drawPixelD( nx1, y, nz, &c  );
	}
}

/* works, but it's too slow for steep lines, because it calls drawPixel.
A better solution is to write drawVLine */
void drawLine ( dFloat x1, dFloat y1, dFloat z1, dFloat x2, dFloat y2,
		dFloat z2, ggi_color *c1, ggi_color *c2 ) {
	fFloat		dx, dz, prevx, x, prevz, z, inv, i;
	VECTOR_UL	start, end;
	ggi_color	prevc, c, dc;
	
	if ( y1 == y2 ) {
		drawHLine ( x1, x2, y1, z1, z2, c1, c2 );
		return;
	}
	
	if ( y1 < y2 ) {
		initV( &start, x1, y1, z1 );	initV( &end, x2, y2, z2 );
		copyColor( &c, c1 );
		inv = 1/(end.y-start.y);
		dc.r = (c2->r - c1->r)*inv;	dc.g = (c2->g - c1->g)*inv;
		dc.b = (c2->b - c1->b)*inv;	dc.a = (c2->a - c1->a)*inv;
	}
	else {
		initV( &end, x1, y1, z1 );	initV( &start, x2, y2, z2 );
		copyColor( &c, c2 );
		inv = 1/(end.y-start.y);
		dc.r = (c1->r - c2->r)*inv;	dc.g = (c1->g - c2->g)*inv;
		dc.b = (c1->b - c2->b)*inv;	dc.a = (c1->a - c2->a)*inv;
	}
	/* initialize */
	x = start.x;			z = start.z;
	
	/* deltas */
	dx = (end.x-start.x)*inv;	dz = (end.z-start.z)*inv;

	for ( i = start.y; i <= end.y; i++ ) { /* for each line */
		prevx = x;		prevz = z;
		copyColor( &prevc, &c );
		x += dx;		z += dz;
		c.r += dc.r;		c.g += dc.g;
		c.b += dc.b;		c.a += dc.a;
		drawHLine( prevx, x, i, prevz, z, &prevc, &c );
	}

}

/* draw a flat triangle. Reasonably well optimized. */
void drawFlatTri ( TRI_UL *tri ) { 
	VECTOR_UL 	*p[3];
	sInt 		a;
	VECTOR_UL	delta1, delta2;
	fFloat		invslope1, invslope2, i, lx, rx, lz, rz, dlz, drz;
	ggi_color	*c;

	/* sort */
	p[0] = MinPoint3( tri->points[0].vect, tri->points[1].vect, tri->points[2].vect );
	p[2] = MaxPoint3( tri->points[0].vect, tri->points[1].vect, tri->points[2].vect );
	for ( a = 0; a < 3; a++ ) {
		if ( p[0] != tri->points[a].vect && p[2] != tri->points[a].vect ) {
			p[1] = tri->points[a].vect;
			break;
		}
	}

	/* deltas for interpolation */
	subV( p[1], p[0], &delta1 );
	subV( p[2], p[0], &delta2 );
	
	/* select color. This is a quick fix to the fact that flatcolor isn't
	well used yet. */
	c = ( tri->flatcolor == NULL ? &tri->color[0] : tri->flatcolor );

	/* first part... */
	if ( delta1.y ) { /* if p[0].x != p[1].x */
		/* slopes */
		invslope1 = (delta1.x/delta1.y);
		invslope2 = (delta2.x/delta2.y);
		/* "differentials" for z buffer interpolation */
		dlz = delta1.z/delta1.y;
		drz = delta2.z/delta2.y;
		/* init right & left x, same for z */
		rx = lx = p[0]->x;
		rz = lz = p[0]->z;
		/* draw the first point */
		drawHLine( lx, lx, p[0]->y, lz, lz, c, c );
		/* now we enter in the loop */
		for ( i = p[0]->y+1; i <= p[1]->y; i++ ) { /* for each line */
			/* if we have some variation */
			if ( (int)(lx += invslope1) != (int)(rx += invslope2) )
				/* draw it */
				drawHLine( lx, rx, i, lz+dlz, rz+drz, c, c );
		}
	}
	else  {/* just draw the line between them */
		drz = delta2.z/delta2.y;
		invslope2 = (delta2.x/delta2.y);
		drawHLine( p[0]->x, p[1]->x, p[0]->y, p[0]->z, p[1]->z, c, c );
	}

	/* corrections for new side */
	subV( p[2], p[1], &delta1 );
	/* now second one... */
	if ( delta1.y ) {
		/* we gotta calculate them again */
		invslope1 = (delta1.x/delta1.y);
		dlz = delta1.z/delta1.y;
		rx = p[1]->x;		rz = p[1]->z;
		lx = p[0]->x + (p[1]->y - p[0]->y) * invslope2;
		lz = p[0]->z + (p[1]->y - p[0]->y) * drz;
		for ( i = p[1]->y+1; i <= p[2]->y; i++ )
			if ( (lx += invslope2 ) != (rx += invslope1) )
				drawHLine( lx, rx, i, lz+dlz, rz+drz, c, c );
	}
	else 
		drawHLine( p[1]->x, p[2]->x, p[0]->y, p[1]->z, p[2]->z, c, c );
}

/* draws a Gouraud triangle. */
/* optimizations: maybe saving 1/delta[123].y and multiplying instead of
dividing will improve things. */
void drawShadedTri ( TRI_UL *tri ) { 
	VECTOR_UL	*p[3], delta1, delta2, delta3;
	sInt 		a;
	fFloat		invslope1, invslope2, i, lx, rx, lz = 0, rz, dlz, drz;
	ggi_color	*c[3], lc, rc, dlc, drc;

	/* sort */
	p[0] = MinPoint3( tri->points[0].vect, tri->points[1].vect, tri->points[2].vect );
	p[2] = MaxPoint3( tri->points[0].vect, tri->points[1].vect, tri->points[2].vect );
	for ( a = 0; a < 3; a++ ) {
		if ( p[0] != tri->points[a].vect && p[2] != tri->points[a].vect ) {
			p[1] = tri->points[a].vect;
			c[1] = &tri->color[a];
			continue;
		}
		else if ( p[0] == tri->points[a].vect ) {
			c[0] = &tri->color[a];
			continue;
		}
		else {
			c[2] = &tri->color[a];
			continue;
		}
	}

	/* deltas for interpolation */
	subV( p[1], p[0], &delta1 );
	subV( p[2], p[0], &delta2 );

	/* first part... */
	if ( delta1.y && delta2.y ) { /* if points do not coincide */
		/* slopes */
		invslope1 = (delta1.x/delta1.y);
		invslope2 = (delta2.x/delta2.y);
		/* "differentials" for z buffer interpolation */
		dlz = delta1.z/delta1.y;
		drz = delta2.z/delta2.y;
		/* for color */
		dlc.r = (c[1]->r - c[0]->r)/delta1.y;
		dlc.g = (c[1]->g - c[0]->g)/delta1.y;
		dlc.b = (c[1]->b - c[0]->b)/delta1.y;
		dlc.a = (c[1]->a - c[0]->a)/delta1.y;
		drc.r = (c[2]->r - c[0]->r)/delta2.y;
		drc.g = (c[2]->g - c[0]->g)/delta2.y;
		drc.b = (c[2]->b - c[0]->b)/delta2.y;
		drc.a = (c[2]->a - c[0]->a)/delta2.y;
		/* init right & left x, same for z and color */
		rx = lx = p[0]->x;
		rz = lz = p[0]->z;
		copyColor( &lc, c[0] );
		copyColor( &rc, c[0] );
		/* draw the first point */
		drawHLine( lx, lx, p[0]->y, lz, lz, c[0], c[0] );
		/* now we enter in the loop */
		for ( i = p[0]->y+1; i <= p[1]->y; i++ ) { /* for each line */
			/* if we have some variation */
			if ( (int)(lx += invslope1) != (int)(rx += invslope2) ) {
				/* draw it */
				lc.r += dlc.r;     rc.r += drc.r;
				lc.g += dlc.g;     rc.g += drc.g;
				lc.b += dlc.b;     rc.b += drc.b;
				lc.a += dlc.a;     rc.a += drc.a;
				drawHLine( lx, rx, i, lz+dlz, rz+drz, &lc, &rc );
			}
		}
	}
	else /* just draw the line between them */
		drawHLine( p[0]->x, p[1]->x, p[0]->y, p[0]->z, p[1]->z, c[0], c[1] );

	/* corrections for new side */
	subV( p[2], p[1], &delta3 );
	/* now second one... */
	if ( delta3.y && delta2.y ) {
		/* we gotta calculate them again */
		invslope1 = (delta3.x/delta3.y);
		invslope2 = (delta2.x/delta2.y);
		dlz = delta3.z/delta3.y;
		drz = delta2.z/delta2.y;
		dlc.r = (c[2]->r - c[1]->r)/delta3.y;
		dlc.g = (c[2]->g - c[1]->g)/delta3.y;
		dlc.b = (c[2]->b - c[1]->b)/delta3.y;
		dlc.a = (c[2]->a - c[1]->a)/delta3.y;
		/* and update */
		rx = p[1]->x;		rz = p[1]->z;
		copyColor( &lc, c[1] );
		lx = p[0]->x + delta1.y * invslope2;
		if ( !delta1.y ) {
			copyColor( &rc, c[0] );
			lz = p[0]->z;
			drc.r = (c[2]->r - c[0]->r)/delta2.y;
			drc.g = (c[2]->g - c[0]->g)/delta2.y;
			drc.b = (c[2]->b - c[0]->b)/delta2.y;
			drc.a = (c[2]->a - c[0]->a)/delta2.y;
		}
		for ( i = p[1]->y+1; i <= p[2]->y; i++ )
			if ( (lx += invslope2 ) != (rx += invslope1) ) {
				/* draw it */
				lc.r += dlc.r;     rc.r += drc.r;
				lc.g += dlc.g;     rc.g += drc.g;
				lc.b += dlc.b;     rc.b += drc.b;
				lc.a += dlc.a;     rc.a += drc.a;
				drawHLine( lx, rx, i, lz+dlz, rz+drz, &rc, &lc );
			}
	}
	else 
		drawHLine( p[1]->x, p[2]->x, p[0]->y, p[1]->z, p[2]->z, c[1], c[2] );
}

void drawTexturedTri ( TRI_UL *tri ) {
	VECTOR_UL	*p[3], delta1, delta2, delta3;
	sInt 		a;
	fFloat		invslope1, invslope2, i, lx, rx, lz = 0, rz, dlz, drz;
	struct { fFloat u, v; } c[3], lc, rc, dlc, drc;
	ggi_color	color;

	/* sort */
	p[0] = MinPoint3( tri->points[0].vect, tri->points[1].vect, tri->points[2].vect );
	p[2] = MaxPoint3( tri->points[0].vect, tri->points[1].vect, tri->points[2].vect );
	for ( a = 0; a < 3; a++ ) {
		if ( p[0] != tri->points[a].vect && p[2] != tri->points[a].vect ) {
			p[1] = tri->points[a].vect;
			c[1].u = tri->points[a].u;
			c[1].v = tri->points[a].v;
			continue;
		}
		else if ( p[0] == tri->points[a].vect ) {
			c[0].u = tri->points[a].u;
			c[0].v = tri->points[a].v;
			continue;
		}
		else {
			c[2].u = tri->points[a].u;
			c[2].v = tri->points[a].v;
			continue;
		}
	}

	/* deltas for interpolation */
	subV( p[1], p[0], &delta1 );
	subV( p[2], p[0], &delta2 );

	/* first part... */
	if ( delta1.y && delta2.y ) { /* if points do not coincide */
		/* slopes */
		invslope1 = (delta1.x/delta1.y);
		invslope2 = (delta2.x/delta2.y);
		/* "differentials" for z buffer interpolation */
		dlz = delta1.z/delta1.y;
		drz = delta2.z/delta2.y;
		/* for color */
		dlc.u = (c[1].u - c[0].u)/delta1.y;
		dlc.v = (c[1].v - c[0].v)/delta1.y;
		drc.u = (c[2].u - c[0].u)/delta2.y;
		drc.v = (c[2].v - c[0].v)/delta2.y;
		/* init right & left x, same for z and color */
		rx = lx = p[0]->x;
		rz = lz = p[0]->z;
		lc.u = rc.u = c[0].u;		lc.v = rc.v = c[0].v;
		/* draw the first point */
		color = pixelImage( tri->texture, c[0].u, c[0].v );
		drawPixelD( lx, p[0]->y, lz, &color );
		/* now we enter in the loop */
		for ( i = p[0]->y+1; i <= p[1]->y; i++ ) { /* for each line */
			/* if we have some variation */
			if ( (int)(lx += invslope1) != (int)(rx += invslope2) ) {
				/* draw it */
				lc.u += dlc.u;     rc.u += drc.u;
				lc.v += dlc.v;     rc.v += drc.v;
				drawHLineTex( lx, rx, i, lz+=dlz, rz+=drz,
						tri->texture, lc.u, rc.u,
						lc.v, rc.v );
			}
		}
	}
	else /* just draw the line between them */
		drawHLineTex( p[0]->x, p[1]->x, p[0]->y, p[0]->z, p[1]->z,
				tri->texture, c[0].u, c[1].u, c[0].v, c[1].v );

	/* corrections for new side */
	subV( p[2], p[1], &delta3 );
	/* now second one... */
	if ( delta3.y && delta2.y ) {
		/* we gotta calculate them again */
		invslope1 = (delta3.x/delta3.y);
		invslope2 = (delta2.x/delta2.y);
		dlz = delta3.z/delta3.y;
		drz = delta2.z/delta2.y;
		dlc.u = (c[2].u - c[1].u)/delta3.y;
		dlc.v = (c[2].v - c[1].v)/delta3.y;
		/* and update */
		rx = p[1]->x;		rz = p[1]->z;
		lc.u = c[1].u;		lc.v = c[1].v;
		lx = p[0]->x + delta1.y * invslope2;
		if ( !delta1.y ) {
			rc.u = c[0].u;		rc.v = c[0].v;
			lz = p[0]->z;
			drc.u = (c[2].u - c[0].u)/delta2.y;
			drc.v = (c[2].v - c[0].v)/delta2.y;
		}
		for ( i = p[1]->y+1; i <= p[2]->y; i++ )
			if ( (lx += invslope2 ) != (rx += invslope1) ) {
				/* draw it */
				lc.u += dlc.u;     rc.u += drc.u;
				lc.v += dlc.v;     rc.v += drc.v;
				drawHLineTex( lx, rx, i, lz+=dlz, rz+=drz,
						tri->texture, rc.u, lc.u,
						rc.v, lc.v );
			}
	}
	else {}
		drawHLineTex( p[1]->x, p[2]->x, p[0]->y, p[1]->z, p[2]->z, 
				tri->texture, c[1].u, c[2].u, c[1].v, c[2].v );
}

/* polygon should be already clipped */
void drawPoly ( POLY_UL *poly ) {
	sInt		i;

	switch ( engpref.gtype ) {
	    case VERTICES:
		for ( i = 0; i < poly->npoints; i++ )
			drawPixelV( poly->points[i].vect,
				&poly->color[(poly->onecolor == True ? 0 : i)] );
		return;
	    case WIRE:
		for ( i = 0; i < poly->npoints; i++ )
			/* this line below is worth entering C Obfuscated...
			 It just draws a line from a vertex to the next one. */
			drawLine( poly->points[i].vect->x,
				poly->points[i].vect->y, poly->points[i].vect->z, 
				poly->points[(i+1 == poly->npoints ? 0 : i+1)].vect->x,
				poly->points[(i+1 == poly->npoints ? 0 : i+1)].vect->y,
				poly->points[(i+1 == poly->npoints ? 0 : i+1)].vect->z,
				(poly->onecolor == True ? (poly->flatcolor == NULL ? 
					&poly->color[0] : poly->flatcolor ) : &poly->color[i]),
				(poly->onecolor == True ? (poly->flatcolor == NULL ? 
					&poly->color[0] : poly->flatcolor ) : &poly->color[(i+1 == poly->npoints ? 0 : i+1)]) );
		fprintf( stderr, "Mode not supported yet" );
		return;

	    case FLAT:
		if ( poly->npoints == 3 ) {
			drawFlatTri( (TRI_UL *)poly );
			return;
		}
		/* poly isn't a triangle let's cut it in triangles */
	
		/* since we're not using u,v or vertex normal we don't copy them */
		ultri.points[0].vect = poly->points[0].vect;
		ultri.points[1].vect = poly->points[1].vect;
		copyColor( ultri.flatcolor, (poly->flatcolor == NULL ? 
			&poly->color[0] : poly->flatcolor) );
		for ( i = 2; i < poly->npoints; i++ ) {
			ultri.points[2].vect = poly->points[i].vect;
			drawFlatTri( &ultri );
		}
		return;

	    case GOURAUD:
		ultri.onecolor = poly->onecolor;
		if ( poly->npoints == 3 ) {
			if ( ultri.onecolor == True )
				drawFlatTri( (TRI_UL *)poly );
			else
				drawShadedTri( (TRI_UL *)poly );
			return;
		}

		/* poly isn't a triangle let's cut it in triangles */
		/* since we're not using u,v or vertex normal we don't copy them */
		ultri.points[0].vect = poly->points[0].vect;
		ultri.points[1].vect = poly->points[1].vect;
		if ( ultri.onecolor == True ) { 
			/* don't waste time and use the flat rasterizer */
			copyColor( ultri.flatcolor, (poly->flatcolor == NULL ? 
				&poly->color[0] : poly->flatcolor) );
			for ( i = 2; i < poly->npoints; i++ ) {
				ultri.points[2].vect = poly->points[i].vect;
				drawFlatTri( &ultri );
			}
		}
		else {
			copyColor( &ultri.color[0], &poly->color[0] );
			copyColor( &ultri.color[1], &poly->color[1] );
			for ( i = 2; i < poly->npoints; i++ ) {
				ultri.points[2].vect = poly->points[i].vect;
				copyColor( &ultri.color[2], &poly->color[i] );
				drawShadedTri( &ultri );
			}
		}
		return;
	    case TEXTURE:
		fprintf( stderr, "Mode not supported yet" );
		return;
	    case TEXTUREGOURAUD:
		fprintf( stderr, "Mode not supported yet" );
		return;
	}
}

#endif
