/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#ifndef ULENGINE_H
#define ULENGINE_H

#include <stdio.h>
#include "mathlib.h"
#include "config.h"

#ifndef CLIENT /* server and none need to know about colors... */
typedef struct { 
	uShort r,g,b,a; 
} ggi_color;
#else
#include <ggi/ggi.h>
#endif

#ifndef OK
# define OK		0
#endif
#ifndef ERROR  
# define ERROR		-1
#endif
#define FATALERROR 	-2

#define SIN( d )	sin( d )
#define COS( d )	cos( d )

#ifdef CLIENT
struct output_ul {
	const ggi_directbuffer	*buf;
	uByte	 		*fbptr;
	ggi_pixelformat		*pix;
	const sInt		*stride;
};
typedef struct output_ul OUTPUT_UL;

struct frame_ul {
	fFloat	 		**zbuf;
	ggi_color		**color;
	uByte			*buffer;
};
typedef struct frame_ul FRAME_UL;

enum graphic_ul {
	VERTICES, WIRE, FLAT, TEXTURE, GOURAUD, TEXTUREGOURAUD
};
typedef enum graphic_ul GRAPHIC_UL;

enum fogtype_ul {
	NONE, LINEAR, EXP, EXP2
};
typedef enum fogtype_ul FOGTYPE_UL;

struct fog_ul {
	FOGTYPE_UL 	type;
	ggi_color 	color;
	dFloat		near, far;	/* until near there's no fog, after 
					   far fog is maximum. if far = 0,
					   no fog. */
};
typedef struct fog_ul FOG_UL;

struct image_ul {
	uByte		*name;
	sInt		width, height;
	ggi_color 	**data;
};
typedef struct image_ul IMAGE_UL;

struct engpref_ul {
	Bool		aa;		/* anti-aliasing? */
	GRAPHIC_UL	gtype;		/* shading type */
	FOG_UL		fog;
};
typedef struct engpref_ul ENGPREF_UL;

struct frustrum_ul {
	PLANE_UL	side[4];
	PLANE_UL	znear;
	PLANE_UL	zfar;
};
typedef struct frustrum_ul FRUSTRUM_UL;

struct camera_ul {
	VECTOR_UL 	pos; 		/* in world coordinates */
	VECTOR_UL	dir;		/* direction of the camera, in world 
					   == X axis */
	dFloat		angle;		/* camera rotation around dir */
	FRUSTRUM_UL	frustrum;	/* camera's frustrum. In camera's
					coordinates. */
	dFloat		znear, zfar;	/* distances of znear and zfar planes
					    if == 0, are not used */
	BASISCHANGE_UL  camtoworld;	/* change coordinate system */
	BASISCHANGE_UL  worldtocam;
};
typedef struct camera_ul CAMERA_UL;

struct parallel_ul {
	VECTOR_UL	dir;	/* direction of light */
	ggi_color	color;		/* color */
	struct parallel_ul *next;
};
typedef struct parallel_ul PARALLEL_UL;

struct spotlight_ul {
	VECTOR_UL	pos; 		/* in world coordinates */
	VECTOR_UL	dir;		/* direction of light, in world */
	dFloat		angle;		/* opening angle */
	ggi_color	color;
	Bool		volumetric;
	struct spotlight_ul *next;
	/* more things, as fallof, etc in the future */
};
typedef struct spotlight_ul SPOTLIGHT_UL;
#endif /* client */

struct transformation_ul {
	VECTOR_UL	translation;
	QUATERNION_UL	rotation;
	QUATERNION_UL	rotationinv;
};
typedef struct transformation_ul TRANSFORMATION_UL;

struct poly_ul {
	uByte		npoints;  	/* no of point in polygon */
	POINT_UL 	*points;
	VECTOR_UL 	normal;
	VECTOR_UL	center;
#ifdef CLIENT
	ggi_color 	*color;		/* one for each vertex */
	ggi_color	*flatcolor;	/* if FLAT is used, this is the 
					   average of all vertices' colors
					   + light info */
	Bool		onecolor;	/* if every point has the same color. */
	IMAGE_UL	*texture;
#endif
	Bool		visible;
};
typedef struct poly_ul TRI_UL;		/* but npoints = 3 */
typedef struct poly_ul POLY_UL;

typedef TRI_UL * TRISTRIP_UL;

struct doodad_ul {
	struct doodad_ul *next;
	uByte		*name;
	sInt		npoly, nvert;
	POLY_UL		**polygons;	/* array of pointers, local coords */
	VECTOR_UL	world;		/* position in world coordinates */
	VECTOR_UL	dir;		/* direction (x axis) in w coord */
	dFloat		angle;		/* rotation around dir, radians */
	/* angle is the angle dood->y make with world->y if dood->x == w->x */
	SPHERE_UL	sphere;		/* circumscribed sphere */
	VECTOR_UL	centroid;	/* in local */
	uInt		ncopies;	/* number of shared copies */
	BASISCHANGE_UL	worldtolocal;	/* change coordinate system */
	BASISCHANGE_UL	localtoworld;
};
typedef struct doodad_ul DOODAD_UL;

/* in transf.c */
extern void transformPoint 	( POINT_UL *p, VECTOR_UL *translation,
					VECTOR_UL *rotation, dFloat angle );
extern void transformPointT	( POINT_UL *p, TRANSFORMATION_UL *t );
extern void transformPoly	( POLY_UL *p, VECTOR_UL *translation,
					VECTOR_UL *rotation, dFloat angle );
extern void transformPolyT	( POLY_UL *p, TRANSFORMATION_UL *t );
extern void transformDoodad	( DOODAD_UL *d, VECTOR_UL *translation,
					VECTOR_UL *rotation, dFloat angle );
extern void transformDoodadT	( DOODAD_UL *d, TRANSFORMATION_UL *t );

/* in geom.c */
extern sInt  readString 	( FILE *fp, uByte *buffer, sInt *size );
extern sInt  readGeometry	( uByte *filename, DOODAD_UL *d );

/* in primitives.c */
extern DOODAD_UL 		*firstdoodad;

extern void pointstoV 		( POINT_UL *p1, POINT_UL *p2, VECTOR_UL *result );
extern void centerP 		( POLY_UL *poly, VECTOR_UL *p );
extern void centroid 		( DOODAD_UL *d, VECTOR_UL *p );
extern SPHERE_UL circSphere 	( DOODAD_UL *d );
#ifdef CLIENT
extern POINT_UL *initPoint 	( dFloat x, dFloat y, dFloat z, dFloat u, 
					dFloat v, dFloat nx, dFloat ny, dFloat nz );
extern POLY_UL *createPolyD 	( sInt n, Bool onecolor, ... );
extern POLY_UL *createPolyP	( sInt n, Bool onecolor, ggi_color *color, 
					POINT_UL *points );
extern POLY_UL *createPolyPV 	( sInt n, Bool onecolor, ggi_color *color, ... );
extern TRISTRIP_UL *createTriStripD ( sInt noftri, ggi_color color, ... );
extern TRISTRIP_UL *createTriStripP ( sInt noftri, ggi_color color, ... );
extern TRISTRIP_UL *createTriStripColorD ( sInt noftri, ... );
extern TRISTRIP_UL *createTriStripColorP ( sInt noftri, ... );
#else
extern POINT_UL *initPoint	( dFloat x, dFloat y, dFloat z, dFloat nx, 
					dFloat ny, dFloat nz );
extern POLY_UL *createPolyD	( sInt n, ... );
extern POLY_UL *createPolyP	( sInt n, POINT_UL *points );
extern POLY_UL *createPolyPV	( sInt n, ... );
extern TRISTRIP_UL *createTriStripD ( sInt noftri, ... );
extern TRISTRIP_UL *createTriStripP ( sInt noftri, ... );
#endif
extern POLY_UL *copyPoly	( POLY_UL *p );
extern void freePoly		( POLY_UL *p );
extern DOODAD_UL *createDoodadP ( uByte *name, POLY_UL **p );
extern DOODAD_UL *createDoodad	( uByte *filename );
extern DOODAD_UL *copyDoodad 	( DOODAD_UL *src );
extern DOODAD_UL *copySharedDoodad ( DOODAD_UL *src );
extern void freeDoodad		( DOODAD_UL *dood );

#ifdef CLIENT
/* in graphic.c */
extern sInt initGraphic		( void );
extern void closeGraphic 	( void );
extern inline void copyColor	( ggi_color *target, ggi_color *src );
extern inline ggi_color RGBAtoColor ( uShort r, uShort g, uShort b, uShort a );
extern void putPixel		( sInt x, sInt y, ggi_color *color );
extern void getPixel		( sInt x, sInt y, ggi_color *color );
extern void drawPixelD		( dFloat x, dFloat y, dFloat z, ggi_color *color );
extern void drawHLine 		( dFloat x1, dFloat x2, dFloat y, dFloat z1,
					dFloat z2, ggi_color *c1, ggi_color *c2 );
extern void drawLine 		( dFloat x1, dFloat y1, dFloat z1, dFloat x2,
					dFloat y2, dFloat z2, ggi_color *c1,
					ggi_color *c2 );
extern void drawFlatTri		( TRI_UL *tri );
extern void drawShadedTri	( TRI_UL *tri );
extern void drawPoly		( POLY_UL *poly );
#define drawPixelV( v, c )	drawPixelD( v->x, v->y, v->z, c )

/* effects.c */
extern void doFog 		( void );
extern ggi_color fogcolor	( void );
#define fogtype()		engpref.fog.type
#define fognear()		engpref.fog.near 
#define fogfar()		engpref.fog.far
extern void setFog 		( FOGTYPE_UL type, ggi_color *color, 
						dFloat near, dFloat far );
/* image.c */
extern IMAGE_UL *loadimage	( const uByte *filename );
extern void freeImage 		( IMAGE_UL *image );
extern ggi_color pixelImage 	( IMAGE_UL *image, fFloat u, fFloat v );

/* init.c */
extern ggi_visual_t		vis;
extern ggi_mode			mode;
extern ggi_pixel		black, white;
extern OUTPUT_UL		im[2];
extern Bool			db;
extern sByte			numframes;
extern ENGPREF_UL		engpref;
extern sInt initULEngine 	( uInt width, uInt height, uInt bits, 
						Bool single, Bool usedb );
extern void endULEngine 	( void );
extern void setGraphics		( GRAPHIC_UL gtype );
extern void setAntialiasing	( Bool setAA );

/* in project.c */
extern uByte			*basis;
extern sInt			framenum;
extern FRAME_UL			frame;
extern sInt initProject 	( sInt width, sInt height );
extern void closeProject 	( void );
extern void chooseCamera 	( CAMERA_UL *cam );
extern void setCameraFrustrum	( void );
extern void setCameraV 		( CAMERA_UL *cam, VECTOR_UL *wpos,
						VECTOR_UL *dir, dFloat angle );
extern void setCameraD 		( CAMERA_UL *cam, dFloat wposx, dFloat wposy, 
						dFloat wposz, dFloat dirx,
						dFloat diry, dFloat dirz,
						dFloat angle );
extern void setCameraZ		( CAMERA_UL *cam, dFloat znear, dFloat zfar );
extern POLY_UL *worldToCameraP	( POLY_UL *poly, Bool clip );
extern POLY_UL *cameraToScreenP	( POLY_UL *poly );
extern void screenToZBufP	( POLY_UL *poly );
extern void projectD		( DOODAD_UL *doodad );
extern sInt render		( CAMERA_UL *cam );
extern sInt putImage		( void );

/* light.c */
extern PARALLEL_UL		*firstparallel;
extern SPOTLIGHT_UL 		*firstspot;
extern PARALLEL_UL *createParallel ( VECTOR_UL *direction, ggi_color color );
extern void setParallel 	( PARALLEL_UL *light, VECTOR_UL *direction, 
						ggi_color color );
extern void freeParallel	( PARALLEL_UL *light );
extern SPOTLIGHT_UL *createSpot ( VECTOR_UL *pos, VECTOR_UL *direction,
						dFloat angle, ggi_color color,
						Bool volumetric );
extern void setSpot 		( SPOTLIGHT_UL *light, VECTOR_UL *pos, 
						VECTOR_UL *direction, dFloat angle,
						ggi_color color, Bool volumetric );
extern void freeSpot		( SPOTLIGHT_UL *light );
extern SPOTLIGHT_UL *copySpot	( SPOTLIGHT_UL *light );


/* update.c */
extern void updateAll 		( Bool sphere, Bool color, Bool normals,
						Bool center );
extern void updateDoodad 	( DOODAD_UL *dood, Bool sphere, Bool color,
						Bool normals, Bool center );
extern void updatePoly 		( POLY_UL *poly, Bool color, Bool normals,
						Bool center );

#endif /* client */

#endif /* ulengine */
